@@ -1,11 +1,13 @@
 use std::fs;
 
-use clap::Shell;
+use clap_complete::{generate_to, Shell};
+use Shell::*;
+//use clap_complete::shells::Shel{Bash, Fish, PowerShell, Elvish};
 
 include!("src/app.rs");
 
 fn main() {
-    let min_version = "1.53";
+    let min_version = "1.56";
 
     match version_check::is_min_version(min_version) {
         Some(true) => {}
@@ -24,7 +26,8 @@ fn main() {
     fs::create_dir_all(&outdir).unwrap();
 
     let mut app = build_app();
-    app.gen_completions("fd", Shell::Bash, &outdir);
-    app.gen_completions("fd", Shell::Fish, &outdir);
-    app.gen_completions("fd", Shell::PowerShell, &outdir);
+    // NOTE: zsh completions are hand written in contrib/completion/_fd
+    for shell in [Bash, PowerShell, Fish, Elvish] {
+        generate_to(shell, &mut app, "fd", &outdir).unwrap();
+    }
 }
